@@ -46,7 +46,9 @@ enum TagId {
 	RSSTagTitle,
 	RSSTagMediaDescription, RSSTagDescription, RSSTagContentEncoded,
 	RSSTagGuid,
+	/* must be defined after GUID, because it can be a link (isPermaLink) */
 	RSSTagLink,
+	RSSTagEnclosure,
 	RSSTagAuthor, RSSTagDccreator,
 	/* Atom */
 	AtomTagUpdated, AtomTagPublished,
@@ -54,6 +56,8 @@ enum TagId {
 	AtomTagMediaDescription, AtomTagSummary, AtomTagContent,
 	AtomTagId,
 	AtomTagLink,
+	AtomTagLinkAlternate,
+	AtomTagLinkEnclosure,
 	AtomTagAuthor,
 	TagLast
 };
@@ -71,7 +75,7 @@ typedef struct field {
 
 enum {
 	FeedFieldTime = 0, FeedFieldTitle, FeedFieldLink, FeedFieldContent,
-	FeedFieldId, FeedFieldAuthor, FeedFieldLast
+	FeedFieldId, FeedFieldAuthor, FeedFieldEnclosure, FeedFieldLast
 };
 
 typedef struct feedcontext {
@@ -96,42 +100,45 @@ static void string_append(String *, const char *, size_t);
 static void string_buffer_realloc(String *, size_t);
 static void string_clear(String *);
 static void string_print_encoded(String *);
+static void string_print_timestamp(String *);
 static void string_print_trimmed(String *);
-static void xml_handler_attr(XMLParser *, const char *, size_t,
-                             const char *, size_t, const char *, size_t);
-static void xml_handler_attr_end(XMLParser *, const char *, size_t,
-                                 const char *, size_t);
-static void xml_handler_attr_start(XMLParser *, const char *, size_t,
-                                   const char *, size_t);
-static void xml_handler_cdata(XMLParser *, const char *, size_t);
-static void xml_handler_data(XMLParser *, const char *, size_t);
-static void xml_handler_data_entity(XMLParser *, const char *, size_t);
-static void xml_handler_end_el(XMLParser *, const char *, size_t, int);
-static void xml_handler_start_el(XMLParser *, const char *, size_t);
-static void xml_handler_start_el_parsed(XMLParser *, const char *,
-                                        size_t, int);
+static void string_print_uri(String *);
+static void xmlattr(XMLParser *, const char *, size_t, const char *, size_t,
+                    const char *, size_t);
+static void xmlattrend(XMLParser *, const char *, size_t, const char *,
+                       size_t);
+static void xmlattrstart(XMLParser *, const char *, size_t, const char *,
+                         size_t);
+static void xmlcdata(XMLParser *, const char *, size_t);
+static void xmldata(XMLParser *, const char *, size_t);
+static void xmldataentity(XMLParser *, const char *, size_t);
+static void xmltagend(XMLParser *, const char *, size_t, int);
+static void xmltagstart(XMLParser *, const char *, size_t);
+static void xmltagstartparsed(XMLParser *, const char *, size_t, int);
 
-/* map tag name to tagid */
-/* RSS, alphabetical order */
+/* map tag name to TagId type */
+/* RSS, must be alphabetical order */
 static FeedTag rsstags[] = {
 	{ STRP("author"),            RSSTagAuthor            },
 	{ STRP("content:encoded"),   RSSTagContentEncoded    },
 	{ STRP("dc:creator"),        RSSTagDccreator         },
 	{ STRP("dc:date"),           RSSTagDcdate            },
 	{ STRP("description"),       RSSTagDescription       },
+	/* RSS: <enclosure url="" />, Atom has <link rel="enclosure" /> */
+	{ STRP("enclosure"),         RSSTagEnclosure         },
 	{ STRP("guid"),              RSSTagGuid              },
 	{ STRP("link"),              RSSTagLink              },
 	{ STRP("media:description"), RSSTagMediaDescription  },
 	{ STRP("pubdate"),           RSSTagPubdate           },
 	{ STRP("title"),             RSSTagTitle             }
 };
-/* Atom, alphabetical order */
+/* Atom, must be alphabetical order */
 static FeedTag atomtags[] = {
 	/* <author><name></name></author> */
 	{ STRP("author"),            AtomTagAuthor           },
 	{ STRP("content"),           AtomTagContent          },
 	{ STRP("id"),                AtomTagId               },
-	/* <link href="" /> */
+	/* Atom: <link href="" />, RSS has <link></link> */
 	{ STRP("link"),              AtomTagLink             },
 	{ STRP("media:description"), AtomTagMediaDescription },
 	{ STRP("published"),         AtomTagPublished        },
@@ -140,8 +147,9 @@ static FeedTag atomtags[] = {
 	{ STRP("updated"),           AtomTagUpdated          }
 };
 
-/* map tagid type to RSS/Atom field */
+/* map TagId type to RSS/Atom field, all tags must be defined */
 static int fieldmap[TagLast] = {
+	[TagUnknown]              = -1,
 	/* RSS */
 	[RSSTagDcdate]            = FeedFieldTime,
 	[RSSTagPubdate]           = FeedFieldTime,
@@ -151,6 +159,7 @@ static int fieldmap[TagLast] = {
 	[RSSTagContentEncoded]    = FeedFieldContent,
 	[RSSTagGuid]              = FeedFieldId,
 	[RSSTagLink]              = FeedFieldLink,
+	[RSSTagEnclosure]         = FeedFieldEnclosure,
 	[RSSTagAuthor]            = FeedFieldAuthor,
 	[RSSTagDccreator]         = FeedFieldAuthor,
 	/* Atom */
@@ -161,7 +170,9 @@ static int fieldmap[TagLast] = {
 	[AtomTagSummary]          = FeedFieldContent,
 	[AtomTagContent]          = FeedFieldContent,
 	[AtomTagId]               = FeedFieldId,
-	[AtomTagLink]             = FeedFieldLink,
+	[AtomTagLink]             = -1,
+	[AtomTagLinkAlternate]    = FeedFieldLink,
+	[AtomTagLinkEnclosure]    = FeedFieldEnclosure,
 	[AtomTagAuthor]           = FeedFieldAuthor
 };
 
@@ -171,34 +182,56 @@ static const char *baseurl = "";
 static FeedContext ctx;
 static XMLParser parser; /* XML parser state */
 
+static String atomlink;
+static enum TagId atomlinktype;
+static int rssidpermalink;
+
+int
+tagcmp(const void *v1, const void *v2)
+{
+	return strcasecmp(((FeedTag *)v1)->name, ((FeedTag *)v2)->name);
+}
+
 /* Unique tagid for parsed tag name. */
 static enum TagId
 gettag(enum FeedType feedtype, const char *name, size_t namelen)
 {
-	size_t i;
+	FeedTag f, *r = NULL;
+
+	f.name = (char *)name;
 
 	switch (feedtype) {
 	case FeedTypeRSS:
-		/* optimization: these are always non-matching */
-		if (namelen < 4 || namelen > 17)
-			return TagUnknown;
-
-		for (i = 0; i < sizeof(rsstags) / sizeof(rsstags[0]); i++)
-			if (istag(rsstags[i].name, rsstags[i].len, name, namelen))
-				return rsstags[i].id;
-		return TagUnknown;
+		r = bsearch(&f, rsstags, sizeof(rsstags) / sizeof(rsstags[0]),
+		        sizeof(rsstags[0]), tagcmp);
+		break;
 	case FeedTypeAtom:
-		/* optimization: these are always non-matching */
-		if (namelen < 2 || namelen > 17)
-			return TagUnknown;
-
-		for (i = 0; i < sizeof(atomtags) / sizeof(atomtags[0]); i++)
-			if (istag(atomtags[i].name, atomtags[i].len, name, namelen))
-				return atomtags[i].id;
-		return TagUnknown;
+		r = bsearch(&f, atomtags, sizeof(atomtags) / sizeof(atomtags[0]),
+		        sizeof(atomtags[0]), tagcmp);
+		break;
 	default:
-		return TagUnknown;
+		break;
 	}
+
+	return r ? r->id : TagUnknown;
+}
+
+static char *
+ltrim(const char *s)
+{
+	for (; *s && isspace((unsigned char)*s); s++)
+		;
+	return (char *)s;
+}
+
+static char *
+rtrim(const char *s)
+{
+	const char *e;
+
+	for (e = s + strlen(s); e > s && isspace((unsigned char)*(e - 1)); e--)
+		;
+	return (char *)e;
 }
 
 /* Clear string only; don't free, prevents unnecessary reallocation. */
@@ -246,12 +279,8 @@ string_print_encoded(String *s)
 	if (!s->data || !s->len)
 		return;
 
-	/* skip leading whitespace */
-	for (p = s->data; *p && isspace((unsigned char)*p); p++)
-		;
-	/* seek location of trailing whitespace */
-	for (e = s->data + s->len; e > p && isspace((unsigned char)*(e - 1)); e--)
-		;
+	p = ltrim(s->data);
+	e = rtrim(p);
 
 	for (; *p && p != e; p++) {
 		switch (*p) {
@@ -272,17 +301,13 @@ string_print_encoded(String *s)
 static void
 string_print_trimmed(String *s)
 {
-	const char *p, *e;
+	char *p, *e;
 
 	if (!s->data || !s->len)
 		return;
 
-	/* skip leading whitespace */
-	for (p = s->data; *p && isspace((unsigned char)*p); p++)
-		;
-	/* seek location of trailing whitespace */
-	for (e = s->data + s->len; e > p && isspace((unsigned char)*(e - 1)); e--)
-		;
+	p = ltrim(s->data);
+	e = rtrim(p);
 
 	for (; *p && p != e; p++) {
 		if (isspace((unsigned char)*p))
@@ -293,6 +318,38 @@ string_print_trimmed(String *s)
 	}
 }
 
+/* always print absolute urls (using global baseurl) */
+void
+string_print_uri(String *s)
+{
+	char link[4096], *p, *e;
+	int c;
+
+	if (!s->data || !s->len)
+		return;
+
+	p = ltrim(s->data);
+	e = rtrim(p);
+	c = *e;
+	*e = '\0';
+	if (absuri(link, sizeof(link), p, baseurl) != -1)
+		fputs(link, stdout);
+	*e = c; /* restore */
+}
+
+/* print as UNIX timestamp, print nothing if the parsed time is invalid */
+void
+string_print_timestamp(String *s)
+{
+	time_t t;
+
+	if (!s->data || !s->len)
+		return;
+
+	if (parsetime(s->data, &t) != -1)
+		printf("%lld", (long long)t);
+}
+
 long long
 datetounix(long long year, int mon, int day, int hour, int min, int sec)
 {
@@ -359,7 +416,6 @@ gettzoffset(const char *s)
 		size_t len;
 		const int offhour;
 	} tzones[] = {
-		{ STRP("A"),   -1 * 3600 },
 		{ STRP("CDT"), -5 * 3600 },
 		{ STRP("CST"), -6 * 3600 },
 		{ STRP("EDT"), -4 * 3600 },
@@ -368,12 +424,9 @@ gettzoffset(const char *s)
 		{ STRP("MST"), -7 * 3600 },
 		{ STRP("PDT"), -7 * 3600 },
 		{ STRP("PST"), -8 * 3600 },
-		{ STRP("M"),   -2 * 3600 },
-		{ STRP("N"),    1 * 3600 },
-		{ STRP("Y"),   12 * 3600 },
 	};
 	const char *p;
-	int tzhour = 0, tzmin = 0;
+	int tzhour = 0, tzmin = 0, c;
 	size_t i, namelen;
 
 	for (; *s && isspace((unsigned char)*s); s++)
@@ -383,7 +436,7 @@ gettzoffset(const char *s)
 	case '+':
 		for (i = 0, p = s + 1; i < 2 && *p && isdigit((unsigned char)*p); i++, p++)
 			tzhour = (tzhour * 10) + (*p - '0');
-		if (*p && !isdigit((unsigned char)*p))
+		if (*p == ':')
 			p++;
 		for (i = 0; i < 2 && *p && isdigit((unsigned char)*p); i++, p++)
 			tzmin = (tzmin * 10) + (*p - '0');
@@ -395,6 +448,23 @@ gettzoffset(const char *s)
 		/* optimization: these are always non-matching */
 		if (namelen < 1 || namelen > 3)
 			return 0;
+		/* ANSI and military zones, defined wrong in RFC822,
+		   see RFC2822 4.3 page 32.
+		   NOTE: zone J is ambiguous (unused/local observed time) and
+		   handled as UTC+0. */
+		if (namelen == 1) {
+			c = (int)*s;
+			/* NOTE: J is not handled: unused / observer local time */
+			if (c >= 'A' && c <= 'I')
+				return c - 'A' + 1; /* +1 through +9 */
+			else if (c >= 'K' && c <= 'M')
+				return c - 'K' + 10; /* 10 through 12 */
+			else if (c >= 'N' && c <= 'Y')
+				return ('Y' - c) - 12; /* -1 through -12 */
+			else
+				return 0;
+		}
+
 		/* compare tz and adjust offset relative to UTC */
 		for (i = 0; i < sizeof(tzones) / sizeof(*tzones); i++) {
 			if (tzones[i].len == namelen &&
@@ -515,24 +585,11 @@ parsetime(const char *s, time_t *tp)
 static void
 printfields(void)
 {
-	char link[4096];
-	time_t t;
-	int r = -1;
-
-	/* parse time, timestamp and formatted timestamp field is empty
-	 * if the parsed time is invalid */
-	if (ctx.fields[FeedFieldTime].str.data)
-		r = parsetime(ctx.fields[FeedFieldTime].str.data, &t);
-	if (r != -1)
-		printf("%lld", (long long)t);
+	string_print_timestamp(&ctx.fields[FeedFieldTime].str);
 	putchar(FieldSeparator);
 	string_print_trimmed(&ctx.fields[FeedFieldTitle].str);
 	putchar(FieldSeparator);
-	/* always print absolute urls */
-	if (ctx.fields[FeedFieldLink].str.data &&
-	    absuri(link, sizeof(link), ctx.fields[FeedFieldLink].str.data,
-	           baseurl) != -1)
-		fputs(link, stdout);
+	string_print_uri(&ctx.fields[FeedFieldLink].str);
 	putchar(FieldSeparator);
 	string_print_encoded(&ctx.fields[FeedFieldContent].str);
 	putchar(FieldSeparator);
@@ -541,6 +598,8 @@ printfields(void)
 	string_print_trimmed(&ctx.fields[FeedFieldId].str);
 	putchar(FieldSeparator);
 	string_print_trimmed(&ctx.fields[FeedFieldAuthor].str);
+	putchar(FieldSeparator);
+	string_print_uri(&ctx.fields[FeedFieldEnclosure].str);
 	putchar('\n');
 }
 
@@ -557,66 +616,91 @@ isattr(const char *name, size_t len, const char *name2, size_t len2)
 }
 
 static void
-xml_handler_attr(XMLParser *p, const char *tag, size_t taglen,
-	const char *name, size_t namelen, const char *value,
-	size_t valuelen)
+xmlattr(XMLParser *p, const char *t, size_t tl, const char *n, size_t nl,
+	const char *v, size_t vl)
 {
 	/* handles transforming inline XML to data */
 	if (ISINCONTENT(ctx)) {
 		if (ctx.contenttype == ContentTypeHTML)
-			xml_handler_data(p, value, valuelen);
+			xmldata(p, v, vl);
 		return;
 	}
 
-	if (ctx.feedtype == FeedTypeAtom) {
-		if (ISCONTENTTAG(ctx)) {
-			if (isattr(name, namelen, STRP("type")) &&
-			   (isattr(value, valuelen, STRP("xhtml")) ||
-			    isattr(value, valuelen, STRP("text/xhtml")) ||
-			    isattr(value, valuelen, STRP("html")) ||
-			    isattr(value, valuelen, STRP("text/html"))))
-			{
+	if (ctx.feedtype == FeedTypeNone)
+		return;
+
+	/* content-type may be: Atom: text, xhtml, html or mime-type.
+	   MRSS (media:description): plain, html. */
+	if (ISCONTENTTAG(ctx)) {
+		if (isattr(n, nl, STRP("type"))) {
+			if (isattr(v, vl, STRP("html")) ||
+			    isattr(v, vl, STRP("xhtml")) ||
+			    isattr(v, vl, STRP("text/html")) ||
+			    isattr(v, vl, STRP("text/xhtml"))) {
 				ctx.contenttype = ContentTypeHTML;
+			} else if (isattr(v, vl, STRP("text")) ||
+			           isattr(v, vl, STRP("plain")) ||
+				   isattr(v, vl, STRP("text/plain"))) {
+				ctx.contenttype = ContentTypePlain;
 			}
+		}
+		return;
+	}
+
+	if (ctx.feedtype == FeedTypeRSS) {
+		if (ctx.tagid == RSSTagEnclosure &&
+		    isattr(n, nl, STRP("url")) && ctx.field) {
+			string_append(ctx.field, v, vl);
+		} else if (ctx.tagid == RSSTagGuid &&
+		           isattr(n, nl, STRP("ispermalink")) &&
+		           !isattr(v, vl, STRP("true"))) {
+			rssidpermalink = 0;
+		}
+	} else if (ctx.feedtype == FeedTypeAtom) {
+		if (ctx.tagid == AtomTagLink &&
+		           isattr(n, nl, STRP("rel"))) {
+			/* empty or "alternate": other types could be
+			   "enclosure", "related", "self" or "via" */
+			if (!vl || isattr(v, vl, STRP("alternate")))
+				atomlinktype = AtomTagLinkAlternate;
+			else if (isattr(v, vl, STRP("enclosure")))
+				atomlinktype = AtomTagLinkEnclosure;
+			else
+				atomlinktype = TagUnknown;
 		} else if (ctx.tagid == AtomTagLink &&
-			   isattr(name, namelen, STRP("href")) &&
-			   ctx.field)
-		{
-			/* link href attribute */
-			string_append(ctx.field, value, valuelen);
+		           isattr(n, nl, STRP("href"))) {
+			string_append(&atomlink, v, vl);
 		}
 	}
 }
 
 static void
-xml_handler_attr_end(XMLParser *p, const char *tag, size_t taglen,
-	const char *name, size_t namelen)
+xmlattrend(XMLParser *p, const char *t, size_t tl, const char *n, size_t nl)
 {
 	if (!ISINCONTENT(ctx) || ctx.contenttype != ContentTypeHTML)
 		return;
 
 	/* handles transforming inline XML to data */
-	xml_handler_data(p, "\"", 1);
+	xmldata(p, "\"", 1);
 	ctx.attrcount = 0;
 }
 
 static void
-xml_handler_attr_start(XMLParser *p, const char *tag, size_t taglen,
-	const char *name, size_t namelen)
+xmlattrstart(XMLParser *p, const char *t, size_t tl, const char *n, size_t nl)
 {
 	if (!ISINCONTENT(ctx) || ctx.contenttype != ContentTypeHTML)
 		return;
 
 	/* handles transforming inline XML to data */
 	if (!ctx.attrcount)
-		xml_handler_data(p, " ", 1);
+		xmldata(p, " ", 1);
 	ctx.attrcount++;
-	xml_handler_data(p, name, namelen);
-	xml_handler_data(p, "=\"", 2);
+	xmldata(p, n, nl);
+	xmldata(p, "=\"", 2);
 }
 
 static void
-xml_handler_cdata(XMLParser *p, const char *s, size_t len)
+xmlcdata(XMLParser *p, const char *s, size_t len)
 {
 	if (!ctx.field)
 		return;
@@ -627,19 +711,19 @@ xml_handler_cdata(XMLParser *p, const char *s, size_t len)
 /* NOTE: this handler can be called multiple times if the data in this
  *       block is bigger than the buffer. */
 static void
-xml_handler_data(XMLParser *p, const char *s, size_t len)
+xmldata(XMLParser *p, const char *s, size_t len)
 {
 	if (!ctx.field)
 		return;
 
 	/* add only data from <name> inside <author> tag
 	 * or any other non-<author> tag */
-	if (ctx.tagid != AtomTagAuthor || istag(p->tag, p->taglen, "name", 4))
+	if (ctx.tagid != AtomTagAuthor || istag(p->tag, p->taglen, STRP("name")))
 		string_append(ctx.field, s, len);
 }
 
 static void
-xml_handler_data_entity(XMLParser *p, const char *data, size_t datalen)
+xmldataentity(XMLParser *p, const char *data, size_t datalen)
 {
 	char buf[16];
 	ssize_t len;
@@ -650,38 +734,31 @@ xml_handler_data_entity(XMLParser *p, const char *data, size_t datalen)
 	/* try to translate entity, else just pass as data to
 	 * xml_data_handler. */
 	if ((len = xml_entitytostr(data, buf, sizeof(buf))) > 0)
-		xml_handler_data(p, buf, (size_t)len);
+		xmldata(p, buf, (size_t)len);
 	else
-		xml_handler_data(p, data, datalen);
+		xmldata(p, data, datalen);
 }
 
 static void
-xml_handler_start_el(XMLParser *p, const char *name, size_t namelen)
+xmltagstart(XMLParser *p, const char *t, size_t tl)
 {
 	enum TagId tagid;
 
 	if (ISINCONTENT(ctx)) {
 		ctx.attrcount = 0;
 		if (ctx.contenttype == ContentTypeHTML) {
-			xml_handler_data(p, "<", 1);
-			xml_handler_data(p, name, namelen);
+			xmldata(p, "<", 1);
+			xmldata(p, t, tl);
 		}
 		return;
 	}
 
 	/* start of RSS or Atom item / entry */
 	if (ctx.feedtype == FeedTypeNone) {
-		if (istag(name, namelen, STRP("entry"))) {
-			/* Atom */
+		if (istag(t, tl, STRP("entry")))
 			ctx.feedtype = FeedTypeAtom;
-			/* default content type for Atom */
-			ctx.contenttype = ContentTypePlain;
-		} else if (istag(name, namelen, STRP("item"))) {
-			/* RSS */
+		else if (istag(t, tl, STRP("item")))
 			ctx.feedtype = FeedTypeRSS;
-			/* default content type for RSS */
-			ctx.contenttype = ContentTypeHTML;
-		}
 		return;
 	}
 
@@ -690,16 +767,40 @@ xml_handler_start_el(XMLParser *p, const char *name, size_t namelen)
 		return;
 
 	/* in item */
-	tagid = gettag(ctx.feedtype, name, namelen);
+	tagid = gettag(ctx.feedtype, t, tl);
 	ctx.tagid = tagid;
 
+	/* without a rel attribute the default link type is "alternate" */
+	if (tagid == AtomTagLink) {
+		atomlinktype = AtomTagLinkAlternate;
+		string_clear(&atomlink); /* reuse and clear temporary link */
+	} else if (tagid == RSSTagGuid) {
+		/* without a ispermalink attribute the default value is "true" */
+		rssidpermalink = 1;
+	}
+
 	/* map tag type to field: unknown or lesser priority is ignored,
 	   when tags of the same type are repeated only the first is used. */
-	if (tagid == TagUnknown || tagid <= ctx.fields[fieldmap[tagid]].tagid) {
+	if (fieldmap[tagid] == -1 || tagid <= ctx.fields[fieldmap[tagid]].tagid) {
 		ctx.field = NULL;
 		return;
 	}
-	ctx.iscontenttag = (fieldmap[ctx.tagid] == FeedFieldContent);
+
+	if (fieldmap[ctx.tagid] == FeedFieldContent) {
+		/* handle default content-type per tag, Atom, RSS, MRSS. */
+		switch (tagid) {
+		case RSSTagContentEncoded:
+		case RSSTagDescription:
+			ctx.contenttype = ContentTypeHTML;
+			break;
+		default:
+			ctx.contenttype = ContentTypePlain;
+		}
+		ctx.iscontenttag = 1;
+	} else {
+		ctx.iscontenttag = 0;
+	}
+
 	ctx.field = &(ctx.fields[fieldmap[ctx.tagid]].str);
 	ctx.fields[fieldmap[ctx.tagid]].tagid = tagid;
 	/* clear field */
@@ -707,8 +808,7 @@ xml_handler_start_el(XMLParser *p, const char *name, size_t namelen)
 }
 
 static void
-xml_handler_start_el_parsed(XMLParser *p, const char *tag, size_t taglen,
-	int isshort)
+xmltagstartparsed(XMLParser *p, const char *tag, size_t taglen, int isshort)
 {
 	if (ctx.iscontenttag) {
 		ctx.iscontent = 1;
@@ -716,17 +816,21 @@ xml_handler_start_el_parsed(XMLParser *p, const char *tag, size_t taglen,
 		return;
 	}
 
+	/* don't read field value in Atom <link> tag */
+	if (ctx.tagid == AtomTagLink)
+		ctx.field = NULL;
+
 	if (!ISINCONTENT(ctx) || ctx.contenttype != ContentTypeHTML)
 		return;
 
 	if (isshort)
-		xml_handler_data(p, "/>", 2);
+		xmldata(p, "/>", 2);
 	else
-		xml_handler_data(p, ">", 1);
+		xmldata(p, ">", 1);
 }
 
 static void
-xml_handler_end_el(XMLParser *p, const char *name, size_t namelen, int isshort)
+xmltagend(XMLParser *p, const char *t, size_t tl, int isshort)
 {
 	size_t i;
 
@@ -735,18 +839,34 @@ xml_handler_end_el(XMLParser *p, const char *name, size_t namelen, int isshort)
 
 	if (ISINCONTENT(ctx)) {
 		/* not close content field */
-		if (gettag(ctx.feedtype, name, namelen) != ctx.tagid) {
+		if (gettag(ctx.feedtype, t, tl) != ctx.tagid) {
 			if (!isshort && ctx.contenttype == ContentTypeHTML) {
-				xml_handler_data(p, "</", 2);
-				xml_handler_data(p, name, namelen);
-				xml_handler_data(p, ">", 1);
+				xmldata(p, "</", 2);
+				xmldata(p, t, tl);
+				xmldata(p, ">", 1);
 			}
 			return;
 		}
+	} else if (ctx.tagid == AtomTagLink) {
+		/* map tag type to field: unknown or lesser priority is ignored,
+		   when tags of the same type are repeated only the first is used. */
+		if (atomlinktype && atomlinktype > ctx.fields[fieldmap[atomlinktype]].tagid) {
+			string_append(&ctx.fields[fieldmap[atomlinktype]].str,
+			              atomlink.data, atomlink.len);
+			ctx.fields[fieldmap[atomlinktype]].tagid = atomlinktype;
+		}
+	} else if (ctx.tagid == RSSTagGuid && rssidpermalink) {
+		if (ctx.tagid > ctx.fields[FeedFieldLink].tagid) {
+			string_clear(&ctx.fields[FeedFieldLink].str);
+			string_append(&ctx.fields[FeedFieldLink].str,
+			             ctx.fields[FeedFieldId].str.data,
+			             ctx.fields[FeedFieldId].str.len);
+			ctx.fields[FeedFieldLink].tagid = ctx.tagid;
+		}
 	} else if (!ctx.tagid && ((ctx.feedtype == FeedTypeAtom &&
-	   istag(name, namelen, STRP("entry"))) || /* Atom */
+	   istag(t, tl, STRP("entry"))) || /* Atom */
 	   (ctx.feedtype == FeedTypeRSS &&
-	   istag(name, namelen, STRP("item"))))) /* RSS */
+	   istag(t, tl, STRP("item"))))) /* RSS */
 	{
 		/* end of RSS or Atom entry / item */
 		printfields();
@@ -757,10 +877,10 @@ xml_handler_end_el(XMLParser *p, const char *name, size_t namelen, int isshort)
 			ctx.fields[i].tagid = TagUnknown;
 		}
 		ctx.contenttype = ContentTypeNone;
-		/* allow parsing of Atom and RSS in one XML stream. */
+		/* allow parsing of Atom and RSS concatenated in one XML stream. */
 		ctx.feedtype = FeedTypeNone;
 	} else if (!ctx.tagid ||
-	           gettag(ctx.feedtype, name, namelen) != ctx.tagid) {
+	           gettag(ctx.feedtype, t, tl) != ctx.tagid) {
 		/* not end of field */
 		return;
 	}
@@ -779,17 +899,17 @@ main(int argc, char *argv[])
 	if (argc > 1)
 		baseurl = argv[1];
 
-	parser.xmlattr = xml_handler_attr;
-	parser.xmlattrend = xml_handler_attr_end;
-	parser.xmlattrstart = xml_handler_attr_start;
-	parser.xmlcdata = xml_handler_cdata;
-	parser.xmldata = xml_handler_data;
-	parser.xmldataentity = xml_handler_data_entity;
-	parser.xmltagend = xml_handler_end_el;
-	parser.xmltagstart = xml_handler_start_el;
-	parser.xmltagstartparsed = xml_handler_start_el_parsed;
+	parser.xmlattr = xmlattr;
+	parser.xmlattrend = xmlattrend;
+	parser.xmlattrstart = xmlattrstart;
+	parser.xmlcdata = xmlcdata;
+	parser.xmldata = xmldata;
+	parser.xmldataentity = xmldataentity;
+	parser.xmltagend = xmltagend;
+	parser.xmltagstart = xmltagstart;
+	parser.xmltagstartparsed = xmltagstartparsed;
 
-	parser.getnext = getchar;
+	/* NOTE: getnext is defined in xml.h for inline optimization */
 	xml_parse(&parser);
 
 	return 0;
